# Written by Bram Cohen
# see LICENSE.txt for license information

from time import time
from cStringIO import StringIO

class DownloaderFeedback:
    def __init__(self, choker, add_task, statusfunc, upfunc, downfunc,
            remainingfunc, leftfunc, file_length, finflag, interval, sp,
            statistics):
        self.choker = choker
        self.add_task = add_task
        self.statusfunc = statusfunc
        self.upfunc = upfunc
        self.downfunc = downfunc
        self.remainingfunc = remainingfunc
        self.leftfunc = leftfunc
        self.file_length = file_length
        self.finflag = finflag
        self.interval = interval
        self.sp = sp
        self.statistics = statistics
        self.lastids = []
        self.display()
        self.spewdata = None

    def _rotate(self):
        cs = self.choker.connections
        for id in self.lastids:
            for i in xrange(len(cs)):
                if cs[i].get_id() == id:
                    return cs[i:] + cs[:i]
        return cs

    def spews(self):
        l = []
        cs = self._rotate()
        self.lastids = [c.get_id() for c in cs]
        for c in cs:
            a = {}
            a['ip'] = c.get_ip()
            if c is self.choker.connections[0]:
                a['optimistic'] = 1
            else:
                a['optimistic'] = 0
            if c.is_locally_initiated():
                a['direction'] = 'l'
            else:
                a['direction'] = 'r'
            u = c.get_upload()
            a['uprate'] = int(u.measure.get_rate())
            if u.is_interested():
                a['uinterested'] = 1
            else:
                a['uinterested'] = 0
            if u.is_choked():
                a['uchoked'] = 1
            else:
                a['uchoked'] = 0
##################################################################
#Modified: Saturday, August 02 2003, 06:14, Karthik
#Purpose: Displaying credit variables.
	    a['optimistic credit']=u.optimistic_credit
	    a['trade credit']=u.trade_credit
	    a['total credit']=u.optimistic_credit + u.trade_credit
# End Modify - Karthik
##################################################################

            d = c.get_download()
            a['downrate'] = int(d.measure.get_rate())
            if d.is_interested():
                a['dinterested'] = 1
            else:
                a['dinterested'] = 1
            if d.is_choked():
                a['dchoked'] = 1
            else:
                a['dchoked'] = 0
            if d.is_snubbed():
                a['snubbed'] = 1
            else:
                a['snubbed'] = 0
            a['utotal'] = d.connection.upload.measure.get_total()
            a['dtotal'] = d.connection.download.measure.get_total()

            a['completed'] = float (float (len(d.connection.download.have)-d.connection.download.unhave)/float(len(d.connection.download.have)))
            a['speed'] = d.connection.download.peermeasure.get_rate()
                                               
            l = l + [a]
        return l

    def spew(self):
        self.spewdata = self.spews()

    def display(self):
        self.add_task(self.display, self.interval)
        self.statistics.update()
        if self.sp:
            self.spew()
        if self.finflag.isSet():
            self.statusfunc(upRate = self.upfunc(),
            statistics = self.statistics, spew = self.spewdata, sizeDone = self.file_length)
            return
        timeEst = self.remainingfunc()

        fractionDone = (self.file_length - self.leftfunc()) / float(self.file_length)
        sizeDone = self.file_length - self.leftfunc()
        
        if timeEst is not None:
            self.statusfunc(timeEst = timeEst, fractionDone = fractionDone, 
                downRate = self.downfunc(), upRate = self.upfunc(),
                statistics = self.statistics, spew = self.spewdata, sizeDone = sizeDone)
        else:
            self.statusfunc(fractionDone = fractionDone, 
                downRate = self.downfunc(), upRate = self.upfunc(),
                statistics = self.statistics, spew = self.spewdata, sizeDone = sizeDone)
