# Written by Edward Keyes
# see LICENSE.txt for license information

from time import time
from copy import copy

class Statistics:
    def __init__(self, upmeasure, downmeasure, connecter):
        self.upmeasure = upmeasure
        self.downmeasure = downmeasure
        self.connecter = connecter
        self.downloader = connecter.downloader
        self.picker = connecter.downloader.picker
        self.torrentmeasure = connecter.downloader.totalmeasure
        self.upTotal = 0.0
        self.downTotal = 0.0
        self.shareRating = 0.0
        self.numSeeds = 0
        self.numOldSeeds = 0
        self.numPeers = 0

    def update(self):
        self.upTotal = self.upmeasure.get_total()
        self.downTotal = self.downmeasure.get_total()
        if self.downTotal > 0:
            self.shareRating = float(self.upTotal)/self.downTotal
        else:
            if self.upTotal == 0:
               self.shareRating = 0.0
            else:
               self.shareRating = -1.0
        self.downloader = self.connecter.downloader
        self.picker = self.downloader.picker
        self.torrentmeasure = self.downloader.totalmeasure
        self.torrentRate = self.torrentmeasure.get_rate()
        self.torrentTotal = self.torrentmeasure.get_total()
        self.numSeeds = 0
        for download in self.downloader.downloads:
            if download.unhave == 0:
                self.numSeeds+=1
        for id in copy(self.downloader.disconnectedseeds.keys()):
            if time()-self.downloader.disconnectedseeds[id]>3600:
               #Expire old seeds after 1 hour from last connection
               del self.downloader.disconnectedseeds[id]
        self.numOldSeeds = len(self.downloader.disconnectedseeds)
        self.numPeers = len(self.downloader.downloads)-self.numSeeds
        self.numCopies = -self.numSeeds
        for i in range(len(self.picker.crosscount)):
            if self.picker.crosscount[i]==0:
                self.numCopies+=1
            else:
                self.numCopies+=1-float(self.picker.crosscount[i])/self.picker.numpieces
                break
        if self.numPeers==0:
            self.percentDone = 0.0
        else:
            self.percentDone = 100.0*(float(self.picker.totalcount)/self.picker.numpieces-self.numSeeds)/self.numPeers


